// ----------------------------------------------------------------------------
// (C) 2002 
//     carrara engineering
//     Switzerland
//     http://www.carrara.ch
//     mailto:info@carrara.ch
// ----------------------------------------------------------------------------
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author make no representations about the suitability 
//     of this software for any purpose. It is provided "as is" 
//     without express or implied warranty.
// ----------------------------------------------------------------------------

#include <iostream.h>

// ----------------------------------------------------------------------------

namespace MetaSingleton1
{

class UserSingleton
{
  public:

    static UserSingleton& instance()
    {
      if (!sm_paSingleton)
      {
        sm_paSingleton = new UserSingleton();
      }
      return *sm_paSingleton;
    }

    const char* getData() const
    {
      return "Singleton's Data";
    }

  private:
    
    UserSingleton() 
    : m_pcData ("Singleton's data...")
    {}

    const char* m_pcData;
    static UserSingleton* sm_paSingleton;
};

UserSingleton* UserSingleton::sm_paSingleton=0;

void test()
{
  cout << UserSingleton::instance().getData() << endl;
}

}

// ----------------------------------------------------------------------------

namespace MetaSingleton2
{

template <class T>
class Singleton
{
  public:

    static T& instance()
    {
      if (!sm_paSingleton)
      {
        sm_paSingleton = new T();
      }
      return *sm_paSingleton;
    }

    static T* sm_paSingleton;

  protected:

    Singleton()
    {
    }

};

template <class T>
T* Singleton<T>::sm_paSingleton=0;

class UserSingleton : public Singleton<UserSingleton>
{
  public:

    const char* getData() const
    {
      return "Singleton's Data";
    }

  private:
    
    friend Singleton<UserSingleton>;
    
    UserSingleton() 
    : m_pcData ("Singleton's data...")
    {}

    const char* m_pcData;
};


void test()
{
  cout << UserSingleton::instance().getData() << endl;
}

}

// ----------------------------------------------------------------------------

namespace MetaSingleton3
{

template <class T>
class Singleton
{
  public:

    Singleton(T& i_raSingleton)
    {
      sm_paSingleton=&i_raSingleton;
    }

    static T& instance()
    {
      if (!sm_paSingleton)
      {
        cout << "Error: Singleton has not been initialized!" << endl;
      }
      return *sm_paSingleton;
    }

  private:

    static T* sm_paSingleton;

};

template <class T>
T* Singleton<T>::sm_paSingleton=0;

class UserData
{
  public:

    UserData(const char* i_pcData) 
    : m_pcData (i_pcData)
    {}

    const char* getData() const
    {
      return m_pcData;
    }

  private:
    
    const char* m_pcData;
};


void f()
{
  cout << Singleton<UserData>::instance().getData() << endl;

}

void test()
{
  UserData aUserData("Singleton's data...");
  Singleton<UserData> aUserDataSingleton(aUserData);

  f();
}

}

// ----------------------------------------------------------------------------

namespace MetaSingleton4
{

template <class T>
class Singleton
{
  public:

    Singleton(T& i_raSingleton)
    : m_bDelete (false)
    {
      sm_paSingleton=&i_raSingleton;
    }

    Singleton(T* i_paSingleton)
    : m_bDelete (true)
    {
      sm_paSingleton=i_paSingleton;
    }

    ~Singleton()
    {
      if (m_bDelete) delete sm_paSingleton;
      sm_paSingleton=0;
    }

    static T& instance()
    {
      if (!sm_paSingleton)
      {
        cout << "Error: Singleton has not been initialized!" << endl;
      }
      return *sm_paSingleton;
    }

  private:

    bool m_bDelete;

    static T* sm_paSingleton;

};

template <class T>
T* Singleton<T>::sm_paSingleton=0;

class UserData1
{
  public:

    UserData1(const char* i_pcData) 
    : m_pcData (i_pcData)
    {}

    const char* getData() const
    {
      return m_pcData;
    }

  private:
    
    const char* m_pcData;
};

class UserData2
{
  public:

    UserData2(const char* i_pcData) 
    : m_pcData (i_pcData)
    {}

    const char* getData() const
    {
      return m_pcData;
    }

  private:
    
    const char* m_pcData;
};


void f()
{
  cout << Singleton<UserData1>::instance().getData() << endl;
  cout << Singleton<UserData2>::instance().getData() << endl;

}

void test()
{
  UserData1 aUserData1("Singleton's data...");
  Singleton<UserData1> aUserData1Singleton(aUserData1);

  Singleton<UserData2> aUserData2Singleton(new UserData2("Singleton's data 2..."));

  f();

}

}

// ----------------------------------------------------------------------------

namespace MetaSingleton5
{

template <class T>
class Singleton
{
  public:

    Singleton(T& i_raSingleton)
    : m_bDelete (false)
    {
      m_paOldSingleton=sm_paSingleton;
      sm_paSingleton=&i_raSingleton;
    }

    Singleton(T* i_paSingleton)
    : m_bDelete (true)
    {
      m_paOldSingleton=sm_paSingleton;
      sm_paSingleton=i_paSingleton;
    }

    ~Singleton()
    {
      if (m_bDelete) delete sm_paSingleton;
      sm_paSingleton=m_paOldSingleton;
    }

    static T& instance()
    {
      if (!sm_paSingleton)
      {
        cout << "Error: Singleton has not been initialized!" << endl;
      }
      return *sm_paSingleton;
    }

  private:

    bool m_bDelete;
    T*   m_paOldSingleton;

    static T* sm_paSingleton;

};

template <class T>
T* Singleton<T>::sm_paSingleton=0;

class UserData
{
  public:

    UserData(const char* i_pcData) 
    : m_pcData (i_pcData)
    {}

    virtual const char* getData() const
    {
      return m_pcData;
    }

  private:
    
    const char* m_pcData;
};

class UserDataDrv : public UserData
{
  public:

    UserDataDrv(const char* i_pcData) 
    : UserData  (i_pcData)
    {}

    virtual const char* getData() const
    {
      const char* pcData = UserData::getData();
      while (*pcData!=' ' && *pcData!=0) ++pcData;
      return *pcData==' '?++pcData:0;
    }

};

void fSub()
{
  cout << Singleton<UserData>::instance().getData() << endl;

}

void f()
{
  cout << Singleton<UserData>::instance().getData() << endl;

  {
    // Reassign
    UserDataDrv aUserDataDrv("Singleton's NEW data...");
    Singleton<UserData> aUserData1Singleton(aUserDataDrv);

    fSub();
  }

  cout << Singleton<UserData>::instance().getData() << endl;

}

void test()
{
  UserData aUserData("Singleton's data...");
  Singleton<UserData> aUserData1Singleton(aUserData);

  f();

}

}

// ----------------------------------------------------------------------------

main()
{
  cout << "--------------------------- Test 1 ---------------------------" << endl;
  MetaSingleton1::test();

  cout << "--------------------------- Test 2 ---------------------------" << endl;
  MetaSingleton2::test();

  cout << "--------------------------- Test 3 ---------------------------" << endl;
  MetaSingleton3::test();

  cout << "--------------------------- Test 4 ---------------------------" << endl;
  MetaSingleton4::test();

  cout << "--------------------------- Test 5 ---------------------------" << endl;
  MetaSingleton5::test();

  return 0;
}
